extern "C" 
{
	#include <wdm.h>
}

struct DEVICE_EXTENSION 
{
	UNICODE_STRING SymLinkName;	// external name
	void* pBuffer;
	ULONG BufferSize;
} ;

// Forward declarations
void DriverUnload ( IN PDRIVER_OBJECT pDrvObj );
NTSTATUS DispatchCreate ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp );
NTSTATUS DispatchClose ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;
NTSTATUS DispatchWrite ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;
NTSTATUS DispatchRead ( IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp ) ;

/*
	Initializes the driver
	Creates the kernel objects needed to process I/O requests
*/
extern "C" NTSTATUS DriverEntry ( IN PDRIVER_OBJECT pDrvObj,
								  IN PUNICODE_STRING pRegPath ) 
{
	// Announce other driver entry points
	pDrvObj->DriverUnload = DriverUnload;
	
	// This includes Dispatch routines for Create, Write & Read
	pDrvObj->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
	pDrvObj->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;
	pDrvObj->MajorFunction[IRP_MJ_READ] = DispatchRead;
	pDrvObj->MajorFunction[IRP_MJ_WRITE] = DispatchWrite;
	
	// Form the internal Device Name
	UNICODE_STRING devname ;
	RtlInitUnicodeString(&devname,L"\\Device\\mydevice"); 

	// Now create the device
	PDEVICE_OBJECT pDevObj ;
	DEVICE_EXTENSION* pDevExt ;
	NTSTATUS status = IoCreateDevice( pDrvObj,
										sizeof(DEVICE_EXTENSION),
										&devname,
										FILE_DEVICE_UNKNOWN,
										0, TRUE, &pDevObj );
	if (!NT_SUCCESS(status))
		return status;

	// Announce that we will be working with a copy of the user's buffer
	pDevObj->Flags |= DO_BUFFERED_IO;

	// Initialize the Device Extension
	pDevExt = (DEVICE_EXTENSION*)pDevObj->DeviceExtension;

	// pDevExt->pDevice = pDevObj;	// back pointer
	// pDevExt->DeviceNumber = ulDeviceNumber;
	// pDevExt->ustrDeviceName = devname;
	pDevExt->pBuffer = NULL;
	pDevExt->BufferSize = 0;

	// Form the symbolic link name
	UNICODE_STRING linkname ;
	RtlInitUnicodeString(&linkname,L"\\??\\mydev");

	pDevExt->SymLinkName = linkname;

	// Now create the link name
	status = IoCreateSymbolicLink( &linkname,&devname );
	if (!NT_SUCCESS(status)) 
	{
		// if it fails now, must delete Device object
		IoDeleteDevice( pDevObj );
		return status;
	}

	return STATUS_SUCCESS;
}

/*
	Stops & Deletes devices controlled by this driver.
	Stops interrupt processing (if any)
	Releases kernel resources consumed by driver
*/
void DriverUnload (IN PDRIVER_OBJECT pDrvObj)
{

	PDEVICE_OBJECT pDevObj = pDrvObj->DeviceObject;
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)pDevObj->DeviceExtension;
	// Free up any buffer still held by this device
	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}
	IoDeleteSymbolicLink(&pDevExt->SymLinkName);
	IoDeleteDevice( pDrvObj->DeviceObject );
	DbgPrint("drv_app1: in driverunload\n");
}

// Handles call from Win32 CreateFile request, does nothing
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
// Return value:
//		NTSTATUS - success or failuer code
//--

NTSTATUS DispatchCreate (IN PDEVICE_OBJECT	pDevObj,IN PIRP pIrp)
{

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

// Function:	DispatchClose
// Description: Handles call from Win32 CreateHandle request, frees any buffer
// Arguments: pDevObj - Passed from I/O Manager pIrp - Passed from I/O Manager
// NTSTATUS - success or failuer code
NTSTATUS DispatchClose (IN PDEVICE_OBJECT	pDevObj,IN PIRP	pIrp)
{
	
	// Dig out the Device Extension from the Device object
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*) pDevObj->DeviceExtension;
	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

// Handles call from Win32 WriteFile request, allocates new pool buffer then xfers user buffer to pool buffer
// Arguments: pDevObj - Passed from I/O Manager pIrp - Passed from I/O Manager
// Return value: NTSTATUS - success or failuer code
NTSTATUS DispatchWrite (IN PDEVICE_OBJECT	pDevObj,IN PIRP pIrp)
{
	NTSTATUS status = STATUS_SUCCESS;
	PVOID userBuffer;
	ULONG xferSize;
	// The stack location contains the user buffer info
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	// Dig out the Device Extension from the Device object
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)	pDevObj->DeviceExtension;
	// Since we processing a new Write request,
	// free up any old buffer
	if (pDevExt->pBuffer != NULL) 
	{
		ExFreePool(pDevExt->pBuffer);
		pDevExt->pBuffer = NULL;
		pDevExt->BufferSize = 0;
	}
	// Determine the length of the request
	xferSize = pIrpStack->Parameters.Write.Length;
	
	// Obtain user buffer pointer
	userBuffer = pIrp->AssociatedIrp.SystemBuffer;

	// Allocate the new buffer
	pDevExt->pBuffer =	ExAllocatePool( PagedPool, xferSize );
	if (pDevExt->pBuffer == NULL) 
	{
		// buffer didn't allocate???
		status = STATUS_INSUFFICIENT_RESOURCES;
		xferSize = 0;
	}
	else 
	{
		// copy the buffer
		pDevExt->BufferSize = xferSize;
		RtlCopyMemory( pDevExt->pBuffer, userBuffer,xferSize );
	}

	// Now complete the IRP
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = xferSize;	// bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return status;
}

// Handles call from Win32 ReadFile request, xfers pool buffer to user
// Arguments: pDevObj - Passed from I/O Manager pIrp - Passed from I/O Manager
// Return value: NTSTATUS - success or failuer code
NTSTATUS DispatchRead (	IN PDEVICE_OBJECT	pDevObj,IN PIRP	pIrp )
{
	NTSTATUS status = STATUS_SUCCESS;

	PVOID userBuffer;
	ULONG xferSize;
	// The stack location contains the user buffer info
	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	
	// Dig out the Device Extension from the Device object
	DEVICE_EXTENSION* pDevExt = (DEVICE_EXTENSION*)	pDevObj->DeviceExtension;
	// Determine the length of the request
	xferSize = pIrpStack->Parameters.Read.Length;
	// Obtain user buffer pointer
	userBuffer = pIrp->AssociatedIrp.SystemBuffer;

	// Don't transfer more than the user's request
	xferSize = (xferSize < pDevExt->BufferSize) ? xferSize : pDevExt->BufferSize;
	
	// Now copy the pool buffer into user space
	RtlCopyMemory( userBuffer, pDevExt->pBuffer,xferSize );
	
	// Free the temporary pool buffer
	ExFreePool( pDevExt->pBuffer );
	pDevExt->pBuffer = NULL;
	pDevExt->BufferSize = 0;

	// Now complete the IRP
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = xferSize;	// bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return status;
}
